/*
  MidiPortImp_Mac.c

  Implementation of native MidiPort methods for Mac
  Author: Robert Marsanyi
  Copyright 1996 Robert Marsanyi
  All Rights Reserved.
*/

/* Generic includes */
#include <StubPreamble.h>
#include "MidiPort.h"

/* OS-related includes */
#include "OMS.h"
#include "OMSDeviceMenu.h"
#if UseMIDIMgr
#include <MIDI.h>
#endif
#include <TextUtils.h>

/* Generic defines */
#define MP_NUM_INPUT_BUFFERS (2)
#define MP_INBUF_SIZE (1024)    /* bytes */
#define MP_CIRCBUF_SIZE (4096)  /* bytes */
/* #define C_TEST */  /* remove for Java library */

/* OS-related defines */
#define BOOLEAN Boolean
#define OUTPORTID 'JMOP'
#define INPORTID  'JMIP'
#define APPSIG    'JVMD'

typedef struct MidiPortContext
{
	char*          mpc_inBuf[MP_NUM_INPUT_BUFFERS];
	char*          mpc_circBuffer;
	volatile char* mpc_circWritePtr, *mpc_circReadPtr;
	BOOLEAN        mpc_BufferOverrun;
	volatile long  mpc_MessagesPending;
	long           mpc_InputErrors;
	BOOLEAN        mpc_enforceNewMessage;
} MidiPortContext;

typedef struct MidiHostContext
{
	OMSUniqueID  mhc_inDeviceID;          /* ID of the input device from which messages are to be received */
	OMSUniqueID  mhc_outDeviceID;         /* ID of the output device to send messages to */
	Boolean      mhc_midiMgrExists;       /* is MIDI Manager present? */
	Boolean      mhc_signedInToMIDIMgr;   /* are we signed into MIDI Manager? */
	Boolean      mhc_signedInToOMS;       /* are we signed into OMS yet? */
	short        mhc_inputPortRefNum;     /* refNum of the local input port */
	short        mhc_outputPortRefNum;    /* refNum of the local output port */
	short	     mhc_compatMode;          /* current OMS compatibility mode */
	OMSNodeInfoListH mhc_inputList;
	OMSNodeInfoListH mhc_outputList;      /* lists of available nodes */
	Boolean      mhc_NotInCallback;
} MidiHostContext;

struct MidiPortContext gMPC;
struct MidiHostContext gMHC;

#ifndef DBUG
#define DBUG(x) /* { printf x; } */
#endif

#define ENABLE_INPUT 1

pascal void MPReceiveMessage( OMSMIDIPacket *pkt, long myRefCon );
long MPStartInput( void );
long MPStopInput( void );

/***************************
Utility functions
***************************/
char *MPPascalToCString( OMSString );
char *MPPascalToCString( OMSString pString )
{
	return( p2cstr( pString ) );
}

/*
MPGetMessageLength
Filched from Bill's M2 code "hr.c"
Args
    statusByte
        MIDI status byte

Results
    length > 1 or Err code < 0.
*/
long MPGetMessageLength (char statusByte);
long MPGetMessageLength (char statusByte)
{
                                          /* 8n 9n An Bn Cn Dn En */
    static const char ChannelLenTable[]  = { 3, 3, 3, 3, 2, 2, 3 };

                                          /* F0 F1 F2 F3 F4 F5 F6 F7 F8 F9 FA FB FC FD FE FF */
    static const char SystemLenTable[]   = { 0, 2, 3, 2, 0, 0, 1, 0, 1, 0, 1, 1, 1, 0, 1, 1 };  /* 0 for undefined (or unknown for F0) */

    long result;

    if (!(statusByte & 0x80)) return -1;        /* !!! error code: bad MIDI status byte */

    result = statusByte < 0xF0
        ? ChannelLenTable [statusByte >> 4 & 0x7]
        : SystemLenTable [statusByte - 0xF0];

    return result > 0 ? result : -1;            /* !!! error code: bad MIDI status byte */
}

/***************************
OMS initialization
***************************/
pascal void MP_AppHook(OMSAppHookMsg *pkt, long myRefCon);
pascal void MP_AppHook(OMSAppHookMsg *pkt, long myRefCon)
{
#ifdef THINK_C
	/*	Set up A5 for access to globals during this routine. */
	asm {
		move.l	a5,-(sp)		; save previous A5
		move.l	myRefCon,a5		; set up A5 from myRefCon
	}
#elif __MWERKS__
	long olda5 = SetA5(myRefCon);
#endif
	
	switch (pkt->msgType) {
	case omsMsgDestDeleted:
		if (gMHC.mhc_outDeviceID == pkt->u.nodeDeleted.uniqueID) {
			gMHC.mhc_outDeviceID = -1;	/* invalid */
		}
		break;
	}

#ifdef THINK_C
	asm {
		move.l	(sp)+,a5		; restore previous A5
	}
#elif __MWERKS__
	SetA5(olda5);
#endif
}

pascal void	MP_ReadHook(OMSMIDIPacket *pkt, long myRefCon);
pascal void	MP_ReadHook(OMSMIDIPacket *pkt, long myRefCon)
{
	short outNodeRefNum;

	/*	Set up A5 for access to globals during this routine. */
#ifdef THINK_C
	asm {
		move.l	a5,-(sp)		; save previous A5
		move.l	myRefCon,a5		; set up A5 from myRefCon
	}
#elif __MWERKS__
	long olda5 = SetA5(myRefCon);
#endif
	
	/*	Process the MIDI packet as efficiently as possible.  It is guaranteed to be MIDI
		data, not some high-level event.  The applications refCon (appRefCon) that was 
		passed to OMSOpenConnection is in the low order word of pkt->tStamp. 
		A convenient way for an application to determine the source of the MIDI data is for 
		it to pass a number identifying the source as the appRefCon to OMSOpenConnection.
		The high-order word of pkt->tStemp is the sources ioRefNum (not its uniqueID); 
		applications can also look at this to determine the source of the MIDI data. */
	
	/*	This sample program merely echoes input to the selected output */
	outNodeRefNum = OMSUniqueIDToRefNum( gMHC.mhc_outDeviceID );
	if (outNodeRefNum != -1)
		OMSWritePacket2(pkt, outNodeRefNum, gMHC.mhc_outputPortRefNum);
	
#ifdef THINK_C
	asm {
		move.l	(sp)+,a5		; restore previous A5
	}
#elif __MWERKS__
	SetA5(olda5);
#endif
}

long MPInitOMS( void );
void MPTermOMS( void );

long MPInitOMS()
{
	OSErr err = 0;
	OMSAppHookUPP appHook;
	short flags;
	
	// First, clean everything out
#if UseMIDIMgr
	gMHC.mhc_midiMgrExists = (SndDispVersion(midiToolNum) != 0);
#else
	gMHC.mhc_midiMgrExists = FALSE;
#endif
	gMHC.mhc_signedInToMIDIMgr = FALSE;
	gMHC.mhc_signedInToOMS = FALSE;
	gMHC.mhc_inputList = NULL;
	gMHC.mhc_outputList = NULL;
	
	/*	Find out what version of OMS is installed, if any. */
	if (OMSVersion() == 0) {
		/*	If OMSVersion returns 0, OMS is not installed. */
		err = -1;
		goto clean;
	}
	
#ifdef powerc
	appHook = NewOMSAppHook(MP_AppHook);
#else
	appHook = MP_AppHook;
#endif
	
	/*	Sign in to OMS */
	err = OMSSignIn(APPSIG, (long)LMGetCurrentA5(), LMGetCurApName(), appHook, &(gMHC.mhc_compatMode));
	/*	Passing CurrentA5 as the refCon solves the problem of A5 setup in the appHook.
		Using other Apple-recommended techniques for setting up A5 in the appHook
		are fine as well.  The client name will be the same as the application's name,
		as stored in the low-memory global CurApName. */
	
	if (err) goto clean;
	else gMHC.mhc_signedInToOMS = TRUE;
	
	// get the OMSNodeInfoList for the input and output separately
	flags = omsIncludeInputs + omsIncludeReal + omsIncludeVirtual + omsIncludeSync;
	if( (gMHC.mhc_inputList = OMSGetNodeInfo(flags)) == NULL )
	{
		err = -1;
		goto clean;
	}
	
	flags = omsIncludeOutputs + omsIncludeReal + omsIncludeVirtual + omsIncludeSync;
	if( (gMHC.mhc_outputList = OMSGetNodeInfo(flags)) == NULL )
	{
		err = -1;
		goto clean;
	}
	
	return( err );

clean:
	MPTermOMS();
	return( err );
}	

void MPTermOMS()
{
	
	if( gMHC.mhc_signedInToOMS )
	{
		if( gMHC.mhc_inputList != NULL )
		{
			OMSDisposeHandle( gMHC.mhc_inputList );
			gMHC.mhc_inputList = NULL;
		}
		if( gMHC.mhc_outputList != NULL )
		{
			OMSDisposeHandle( gMHC.mhc_outputList );
			gMHC.mhc_outputList = NULL;
		}
		OMSSignOut(APPSIG);
	}
}

/***************************
Midi GetNumDevices, GetDeviceName Routines
***************************/

long MidiPort_nGetNumDevices(struct HMidiPort *me,long type)
{
	long err = -1;
	
	// make sure we've inited OMS
	if( !gMHC.mhc_signedInToOMS )
	{
		if( (err = MPInitOMS()) != 0 )
			goto clean;
	}
	
	// get the OMSNodeInfoList for the specified type
	switch( type )
	{
	case MidiPort_MIDIPORT_INPUT:
		if( gMHC.mhc_inputList != NULL )
			return( (*gMHC.mhc_inputList)->numNodes );
		break;
	case MidiPort_MIDIPORT_OUTPUT:
		if( gMHC.mhc_outputList != NULL )
			return( (*gMHC.mhc_outputList)->numNodes );
		break;
	}

clean:	
	return( err );
}

struct Hjava_lang_String *MidiPort_nGetDeviceName(struct HMidiPort *me,long type,long num)
{
	char *name;
	Hjava_lang_String *jString = NULL;
	OMSNodeInfoRec *deviceRec;
	
	// make sure we've inited OMS
	if( !gMHC.mhc_signedInToOMS )
	{
		if( MPInitOMS() != 0 )
			goto clean;
	}
	
	switch( type )
	{
		case MidiPort_MIDIPORT_INPUT:
			if( num < 0 || (num > (*gMHC.mhc_inputList)->numNodes - 1) )
				goto clean;
			else
			{
				deviceRec = &((*gMHC.mhc_inputList)->info[num]);
				name = MPPascalToCString( deviceRec->name );
			
				// make a Java String from the name
				jString = makeJavaString( name, strlen(name) );
			}
			break;
		case MidiPort_MIDIPORT_OUTPUT:
			if( num < 0 || (num > (*gMHC.mhc_outputList)->numNodes - 1) )
				goto clean;
			else
			{
				deviceRec = &((*gMHC.mhc_outputList)->info[num]);
				name = MPPascalToCString( deviceRec->name );
				
				// make a Java String from the name
				jString = makeJavaString( name, strlen(name) );
			}
			break;
	}

clean:
	return(jString);
}

/***************************
Midi Open, Close Routines
***************************/

long MidiPort_nOpen(struct HMidiPort *me)
{
	OMSReadHookUPP readHook;
	OSErr err;
	int i;
	OMSNodeInfoRec *deviceRec;
	long inIndex, outIndex;
	
	// sign into OMS, if you haven't already
	if( !gMHC.mhc_signedInToOMS )
	{
		if( (err = MPInitOMS()) != 0 )
			goto clean;
	}
		
	// extract the requested device from the HMidiPort structure
	#ifdef C_TEST
		outIndex = 0;
		inIndex = 0;
	#else
		outIndex = unhand(me)->outputDeviceNumber;
		inIndex = unhand(me)->inputDeviceNumber;
	#endif
	
	/*  Save inDevice, outDevice */
	if( inIndex >= 0 && inIndex < (*gMHC.mhc_inputList)->numNodes)
	{
		deviceRec = &((*gMHC.mhc_inputList)->info[inIndex]);
		gMHC.mhc_inDeviceID = deviceRec->uniqueID;
	}
	else
	{
		err = -1;
		goto clean;
	}
	
	if( outIndex >= 0 && outIndex < (*gMHC.mhc_outputList)->numNodes)
	{
		deviceRec = &((*gMHC.mhc_outputList)->info[outIndex]);
		gMHC.mhc_outDeviceID = deviceRec->uniqueID;
	}
	else
	{
		err = -1;
		goto clean;
	}
	
	// allocate buffers for sysex input
	for( i = 0; i < MP_NUM_INPUT_BUFFERS; i++ )
	{
		if( (gMPC.mpc_inBuf[i] = malloc( MP_INBUF_SIZE )) == NULL )
		{
			err = -1;
			goto clean;
		}
	}

	// allocate a circular buffer for all input messages
	if( (gMPC.mpc_circBuffer = malloc( MP_CIRCBUF_SIZE )) == NULL )
	{
		err = -1;
		goto clean;
	}

#ifdef powerc
	readHook = NewOMSReadHook(MPReceiveMessage);
#else
	readHook = MPReceiveMessage;
#endif
	
#ifdef ENABLE_INPUT
	/*	Add an input port */
	err = OMSAddPort(APPSIG, INPORTID, omsPortTypeInput2, (OMSReadHook2UPP)readHook, 
		(long)LMGetCurrentA5(), &(gMHC.mhc_inputPortRefNum));
	if (err) goto clean;
	
	err = MPStartInput();
	if( err ) goto clean;
#endif
	
	/*	Add an output port */
	err = OMSAddPort(APPSIG, OUTPORTID, omsPortTypeOutput, NULL, 0L,
		 &(gMHC.mhc_outputPortRefNum));
	if (err) goto clean;
	
	return err;

clean:
	MidiPort_nClose( me );
	return( err );
}

long MidiPort_nClose(struct HMidiPort *me)
{
	long err = 0;
	int i;
	
#ifdef ENABLE_INPUT
	MPStopInput();
#endif
	
	if( gMHC.mhc_signedInToOMS )
	{
		MPTermOMS();
	}

	if( gMPC.mpc_circBuffer != NULL )
	{
		free( gMPC.mpc_circBuffer );
		gMPC.mpc_circBuffer = NULL;
	}

	for( i = 0; i < MP_NUM_INPUT_BUFFERS; i++ )
	{
		if( gMPC.mpc_inBuf[i] != NULL )
		{
			free( gMPC.mpc_inBuf[i] );
			gMPC.mpc_inBuf[i] = NULL;
		}
	}

	return( 0 );
}

/***************************
Midi Output Routines
***************************/

long MidiPort_nWriteShortMessage(struct HMidiPort *me, long status, long data1, long data2)
{
	OMSMIDIPacket pkt;
	short outNodeRefNum;
	
	outNodeRefNum = OMSUniqueIDToRefNum( gMHC.mhc_outDeviceID );
	if( outNodeRefNum == OMSInvalidRefNum )
		return( -1 );
	else
	{
		pkt.flags = 0;
		pkt.len = MPGetMessageLength( status );
		pkt.data[0] = status;	// status byte
		pkt.data[1] = data1;
		pkt.data[2] = data2;
		OMSWritePacket2(&pkt, outNodeRefNum, gMHC.mhc_outputPortRefNum);
		return( 0 );  // OMSWritePacket doesn't return an error code!
	}
}

long MidiPort_nWriteLongMessage(struct HMidiPort *me,HArrayOfByte *buf,long len,long timeout)
{
	long err = 0;
	
	OMSMIDIPacket255 pkt;
	long packetLength, numLeft;
	char* packetPtr;
	short outNodeRefNum;

	outNodeRefNum = OMSUniqueIDToRefNum( gMHC.mhc_outDeviceID );
	if( outNodeRefNum == OMSInvalidRefNum )
	{
		err = OMSInvalidRefNum;
		goto clean;
	}
	else
	{
		numLeft = len;
		packetPtr = unhand(buf)->body;
		
		while( numLeft > 0 )
		{
			// figure out how much to copy
			packetLength = ( numLeft <= 255 ) ? numLeft : 255;
			
			// copy a portion into the OMSMIDIPacket255
			memcpy( pkt.data, packetPtr, packetLength );
			
			pkt.flags = 0;
			pkt.len = packetLength;
			
			// if it's not the first packet, set the previous flag
			if( numLeft != len ) pkt.flags |= omsStartCont;
			
			// advance the pointer
			packetPtr += packetLength;
			numLeft -= packetLength;

			// if it's not the last packet, set the next flag
			if( numLeft > 0 ) pkt.flags |= omsEndCont;
			
			// send the packet
			OMSWritePacket255(&pkt, outNodeRefNum, gMHC.mhc_outputPortRefNum);	
		}
	}
	
clean:
	return( err );
}

/***************************
Midi Input Routines
***************************/
#ifdef ENABLE_INPUT

/*
MPStartInput
sets the read and write pointers to the start of the buffer, queues things
*/

long MPStartInput( void )
{
	long err = 0;
	int i;
	OMSConnectionParams connectionParams;

	gMPC.mpc_circReadPtr = gMPC.mpc_circWritePtr = gMPC.mpc_circBuffer;
	gMPC.mpc_enforceNewMessage = TRUE;

	// set the connection paramater structure up
	connectionParams.nodeUniqueID = gMHC.mhc_inDeviceID;
	connectionParams.appRefCon = 0;
	
	// start processing MIDI input
	gMHC.mhc_NotInCallback = TRUE;
	err = OMSOpenConnections(APPSIG, INPORTID, 1, &connectionParams, FALSE);

	gMPC.mpc_BufferOverrun = FALSE;

clean:
	return(err);
}

long MPStopInput( void )
{
	long err = 0;
	OMSConnectionParams connectionParams;
	
	// set the connection paramater structure up
	connectionParams.nodeUniqueID = gMHC.mhc_inDeviceID;
	connectionParams.appRefCon = 0;
	
	// stop processing MIDI input
	OMSCloseConnections( APPSIG, INPORTID, 1, &connectionParams );
	gMPC.mpc_MessagesPending = 0;

clean:
	return( err );
}

/*
MPCircBufferFree
returns the number of bytes in the buffer that are free
*/
unsigned long MPCircBufferFree( void );
unsigned long MPCircBufferFree( void )
{
	long distance = (long)gMPC.mpc_circReadPtr - (long)gMPC.mpc_circWritePtr;
	if( distance <= 0 ) distance += MP_CIRCBUF_SIZE;
	return((unsigned)distance);
}

/*
MPWriteToInputBuffer
Write a variable number of bytes to the circular buffer, if possible
*/
int MPWriteToInputBuffer( unsigned char* data, unsigned long size );
int MPWriteToInputBuffer( unsigned char* data, unsigned long size )
{
	int overrun = 0;

	if( MPCircBufferFree() >= size )
	{
		// figure length(s) of chunks to copy
		unsigned long size1, size2;
		long dwToEnd = (((char*)gMPC.mpc_circBuffer + MP_CIRCBUF_SIZE) - gMPC.mpc_circWritePtr);

		if( size > dwToEnd )
		{
			size1 = dwToEnd;
			size2 = size - size1;
		}
		else
		{
			size1 = size;
			size2 = 0;
		}

		// copy the first and (optionally) the second chunk to the circular buffer
		memcpy( gMPC.mpc_circWritePtr, data, size1 );
		if( size2 > 0 ) memcpy( gMPC.mpc_circBuffer, data + size1, size2 );
	
		// advance the write pointer
		if( size2 == 0 ) gMPC.mpc_circWritePtr += size1;
		else gMPC.mpc_circWritePtr = gMPC.mpc_circBuffer + size2;

	}
	else
	{
		overrun = -1;
	}

	return( overrun );
}

/*
MPReceiveMessage
This is the callback function for the MIDI device.  It basically queues messages in the circular
buffer for subsequent reading by the client app.
*/
pascal void MPReceiveMessage( OMSMIDIPacket *pkt, long myRefCon )
{
	int err = 0;

	/*	Set up A5 for access to globals during this routine. */
#ifdef THINK_C
	asm {
		move.l	a5,-(sp)		; save previous A5
		move.l	myRefCon,a5		; set up A5 from myRefCon
	}
#elif __MWERKS__
	long olda5 = SetA5(myRefCon);
#endif

	if( !gMPC.mpc_BufferOverrun && gMHC.mhc_NotInCallback )
	{
		gMHC.mhc_NotInCallback = FALSE;
		if( pkt->len > 0 )
		{
			if( ((gMPC.mpc_enforceNewMessage == TRUE) && !(pkt->flags & omsEndCont))
			  || (gMPC.mpc_enforceNewMessage == FALSE) )
			{
				if( MPWriteToInputBuffer( pkt->data, pkt->len ) < 0 )
				{
					gMPC.mpc_BufferOverrun = TRUE;
				}
				else
				{
					if( !( pkt->flags & omsStartCont ) )
					{
						gMPC.mpc_MessagesPending += 1;
						gMPC.mpc_enforceNewMessage = TRUE;
					}
					else
					{
						gMPC.mpc_enforceNewMessage = FALSE;
					}
				}
			}
		}
		gMHC.mhc_NotInCallback = TRUE;
	}
#ifdef THINK_C
	asm {
		move.l	(sp)+,a5		; restore previous A5
	}
#elif __MWERKS__
	SetA5(olda5);
#endif
}

long MidiPort_nReadMessage(struct HMidiPort *this,HArrayOfByte *buf,long len)
{
	long numRead = 0;

	DBUG(("Reading: old (rd,wr) = (0x%x, 0x%x)\n", gMPC.mpc_circReadPtr, gMPC.mpc_circWritePtr));

	// if there's a message waiting
	if( (gMPC.mpc_circReadPtr != gMPC.mpc_circWritePtr) && (gMPC.mpc_MessagesPending > 0) )
	{
		char* lpByteArray = unhand(buf)->body;

		// read up to the next status byte, or until buffer is empty, or until the destination is full
		do
		{
			lpByteArray[numRead++] = *gMPC.mpc_circReadPtr++;
			if( gMPC.mpc_circReadPtr >= gMPC.mpc_circBuffer + MP_CIRCBUF_SIZE )
				gMPC.mpc_circReadPtr = gMPC.mpc_circBuffer;
		}
		while( !(*gMPC.mpc_circReadPtr & 0x80) && (numRead < len) && (gMPC.mpc_circReadPtr != gMPC.mpc_circWritePtr));

		// include end sysex, if it's there
		if( ((*gMPC.mpc_circReadPtr & 0xFF) == 0xF7) && (numRead < len) && (gMPC.mpc_circReadPtr != gMPC.mpc_circWritePtr))
		{
			lpByteArray[numRead++] = *gMPC.mpc_circReadPtr++;
			if( gMPC.mpc_circReadPtr >= gMPC.mpc_circBuffer + MP_CIRCBUF_SIZE )
				gMPC.mpc_circReadPtr = gMPC.mpc_circBuffer;
		}

		// if the message is complete, decrement messagesPending
		if( (lpByteArray[numRead - 1] == 0xF7) || (gMPC.mpc_circReadPtr == gMPC.mpc_circWritePtr) || (*gMPC.mpc_circReadPtr & 0x80) )
			gMPC.mpc_MessagesPending--;
	}

	DBUG(("Reading: new (rd,wr,pending) = (0x%x, 0x%x, %i)\n", gMPC.mpc_circReadPtr, gMPC.mpc_circWritePtr,
	  gMPC.mpc_MessagesPending));

	// return number of bytes read
	return( numRead );
}

long MidiPort_nMessagesWaiting(struct HMidiPort *this)
{
	if( gMPC.mpc_BufferOverrun == TRUE ) return( -1 );
	else return( gMPC.mpc_MessagesPending );
}

long MidiPort_nResetInput(struct HMidiPort *this)
{
	long err = 0;
	
	if( (err = MPStopInput()) != 0 )
		goto clean;
	if( (err = MPStartInput()) != 0 )
		goto clean;

clean:
	return(err);
}

#else

long MidiPort_nReadMessage(struct HMidiPort *this,HArrayOfByte *buf,long len)
{
	return(-1);
}

long MidiPort_nMessagesWaiting(struct HMidiPort *this)
{
	return(-1);
}

long MidiPort_nResetInput(struct HMidiPort *this)
{
	return(-1);
}

#endif